use {
    crate::{error::Error, B64Alphabet},
    std::io::{ErrorKind, Read, Write},
};

/// Decodes b64 encoded data
/// # Exammple
/// ```
/// # use {
/// #    std::io::Read,
/// #    b64::Decoder,
/// # };
/// #
/// # fn main() {
/// let mut decoder = Decoder::new("SGVsbG8sIFdvcmxkIQ==".as_bytes(), vec![], None, false);
/// assert_eq!(decoder.decode().unwrap(), b"Hello, World!");
/// # }
/// ```
pub struct Decoder<R: Read, W: Write> {
    reader: R,
    writer: W,
    alphabet: B64Alphabet,
    ignore_whitespace: bool,
}

impl<R: Read, W: Write> Decoder<R, W> {
    pub fn new(
        reader: R,
        writer: W,
        alphabet: Option<B64Alphabet>,
        ignore_whitespace: bool,
    ) -> Self {
        Self {
            reader,
            writer,
            alphabet: alphabet.unwrap_or_default(),
            ignore_whitespace,
        }
    }

    /// Decodes the bytes provided by `self.reader` and writes the resulting
    /// bytes into `self.writer`.
    /// # Important
    /// Both reader and writer should be buffered for good performance, as the
    /// process of decoding will perform a series of very short read and write
    /// operations.
    /// # Errors
    /// - io failure will return `Error::Io`
    /// - a character not in the alphabet, or encountering any character after
    ///   the first occurance of the padding character, will return `Error::IllegalChar`
    /// - if the reader comes up short of the final four byte block will return
    ///   `Error::MissingPadding`
    /// - `intError` should never be returned. If this error is recieved the code
    ///   is somehow broken. Please file a bug.
    pub fn decode(mut self) -> Result<W, Error> {
        let mut byte_reader = self.reader.bytes();
        loop {
            let mut in_buf = [0_u8; 4];
            let mut out_buf = [0_u8; 3];
            let mut num = 0_u32;
            let mut n_bytes = 0;
            while n_bytes < 4 {
                match byte_reader.next() {
                    Some(Ok(b)) => {
                        if self.ignore_whitespace && b.is_ascii_whitespace() {
                            continue;
                        } else if b == b'\n' || b == b'\r' {
                            continue;
                        } else {
                            in_buf[n_bytes] = b;
                            n_bytes += 1;
                        }
                    }
                    Some(Err(e)) if e.kind() == ErrorKind::UnexpectedEof => break,
                    Some(Err(e)) if e.kind() == ErrorKind::Interrupted => continue,
                    Some(Err(e)) => return Err(e.into()),
                    None => break,
                }
            }
            match n_bytes {
                0 => break,
                4 => {}
                _ => return Err(Error::MissingPadding),
            }
            let mut bytes = 0;
            for (i, &c) in in_buf.iter().enumerate() {
                num <<= 6;
                if c == self.alphabet.pad() {
                    continue;
                }
                if i != bytes {
                    return Err(Error::IllegalChar(c.into()));
                }
                let Some(idx) = self.alphabet.get(c) else {
                    return Err(Error::IllegalChar(c.into()));
                };
                num |= u32::try_from(idx)?;
                bytes += 1;
            }
            let olen = bytes * 6 / 8;
            for i in (0..3).rev() {
                out_buf[i] = (num & 0xff) as u8;
                num >>= 8;
            }
            self.writer.write_all(&out_buf[0..olen])?;
        }
        Ok(self.writer)
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use std::{
        fs::File,
        io::{BufReader, Read},
    };

    #[test]
    fn decode() {
        let mut decoder = Decoder::new("SGVsbG8sIFdvcmxk".as_bytes(), vec![], None, false);
        assert_eq!(decoder.decode().unwrap(), b"Hello, World");
        decoder = Decoder::new("SGVsbG8sIFdvcmxkIQ==".as_bytes(), vec![], None, false);
        assert_eq!(decoder.decode().unwrap(), b"Hello, World!");
        decoder = Decoder::new("SGVsbG8sIFdvcmxkIQo=".as_bytes(), vec![], None, false);
        assert_eq!(decoder.decode().unwrap(), b"Hello, World!\n");
    }

    #[test]
    fn ignore_whitespace() {
        let decoder = Decoder::new(" SG\tVsbG8sIFdvcmxkIQ==".as_bytes(), vec![], None, true);
        assert_eq!(decoder.decode().unwrap(), b"Hello, World!");
    }

    #[test]
    fn decode_wrapped() {
        let infile = File::open("testdata/lorem_b64.txt").unwrap();
        let reader = BufReader::new(infile);
        let mut outfile = vec![];
        File::open("testdata/lorem.txt")
            .unwrap()
            .read_to_end(&mut outfile)
            .unwrap();
        let decoder = Decoder::new(reader, vec![], None, false);
        assert_eq!(decoder.decode().unwrap(), outfile);
    }
}
